#include <windows.h>

BOOL SetPrinterProperties(LPTSTR pPrinterName, DEVMODE *pDevModeNew)
{
	HANDLE hPrinter = NULL;
	DWORD dwNeeded = 0;
	PRINTER_INFO_2 *pi2 = NULL;
	DEVMODE *pDevMode = NULL;
	PRINTER_DEFAULTS pd;
	BOOL bFlag;
	LONG lFlag;

	// Open printer handle 
	ZeroMemory(&pd, sizeof(pd));
	pd.DesiredAccess = PRINTER_ACCESS_USE; //PRINTER_ALL_ACCESS;
	bFlag = OpenPrinter(pPrinterName, &hPrinter, &pd);
	if (!bFlag || (hPrinter == NULL))
		return FALSE;

	// The first GetPrinter tells you how big the buffer should be in
	// order to hold all of PRINTER_INFO_2. Note that this should fail with
	// ERROR_INSUFFICIENT_BUFFER.  If GetPrinter fails for any other reason
	// or dwNeeded isn't set for some reason, then there is a problem...
	SetLastError(0);
	bFlag = GetPrinter(hPrinter, 2, 0, 0, &dwNeeded);
         if ((!bFlag) && (GetLastError() != ERROR_INSUFFICIENT_BUFFER) ||
         (dwNeeded == 0))
	{
		ClosePrinter(hPrinter);
		return FALSE;
	}

	// Allocate enough space for PRINTER_INFO_2...
	pi2 = (PRINTER_INFO_2 *)GlobalAlloc(GPTR, dwNeeded);
	if (pi2 == NULL)
	{
		ClosePrinter(hPrinter);
		return FALSE;
	}

	// The second GetPrinter fills in all the current settings, so all you
	// need to do is modify what you're interested in...
	bFlag = GetPrinter(hPrinter, 2, (LPBYTE)pi2, dwNeeded, &dwNeeded);
	if (!bFlag)
	{
		GlobalFree(pi2);
		ClosePrinter(hPrinter);
		return FALSE;
	}

	// If GetPrinter didn't fill in the DEVMODE, try to get it by calling
	// DocumentProperties...
	if (pi2->pDevMode == NULL)
	{
		dwNeeded = DocumentProperties(NULL, hPrinter,
						pPrinterName,
						NULL, NULL, 0);
		if (dwNeeded <= 0)
		{
			GlobalFree(pi2);
			ClosePrinter(hPrinter);
			return FALSE;
		}

		pDevMode = (DEVMODE *)GlobalAlloc(GPTR, dwNeeded);
		if (pDevMode == NULL)
		{
			GlobalFree(pi2);
			ClosePrinter(hPrinter);
			return FALSE;
		}

		lFlag = DocumentProperties(NULL, hPrinter,
						pPrinterName,
						pDevMode, NULL,
						DM_OUT_BUFFER);

		if (lFlag != IDOK || pDevMode == NULL)
		{
			GlobalFree(pDevMode);
			GlobalFree(pi2);
			ClosePrinter(hPrinter);
			return FALSE;
		}

		pi2->pDevMode = pDevMode;
	}

	if (pDevModeNew->dmFields & DM_COPIES) {
		if(pi2->pDevMode->dmCopies != pDevModeNew->dmCopies) {
			// Specify exactly what we are attempting to change...
			pi2->pDevMode->dmFields |= DM_COPIES;
			pi2->pDevMode->dmCopies = pDevModeNew->dmCopies;
		}
	}
	
	if (pDevModeNew->dmFields & DM_ORIENTATION) {
		if(pi2->pDevMode->dmOrientation != pDevModeNew->dmOrientation ) {
			// Specify exactly what we are attempting to change...
			pi2->pDevMode->dmFields |= DM_ORIENTATION;
			pi2->pDevMode->dmOrientation = pDevModeNew->dmOrientation;
		}
	}

	if (pDevModeNew->dmFields & DM_DUPLEX) {
		if(pi2->pDevMode->dmDuplex != pDevModeNew->dmDuplex) {
			// Specify exactly what we are attempting to change...
			pi2->pDevMode->dmFields |= DM_DUPLEX;
			pi2->pDevMode->dmDuplex = pDevModeNew->dmDuplex;
		}
	}

	if (pDevModeNew->dmFields & DM_SCALE) {
		if(pi2->pDevMode->dmScale != pDevModeNew->dmScale) {
			// Specify exactly what we are attempting to change...
			pi2->pDevMode->dmFields |= DM_SCALE;
			pi2->pDevMode->dmNup = DMNUP_SYSTEM;
			pi2->pDevMode->dmScale = pDevModeNew->dmScale;
		}
	}


	// Do not attempt to set security descriptor...
	pi2->pSecurityDescriptor = NULL;

	// Make sure the driver-dependent part of devmode is updated...
	lFlag = DocumentProperties(NULL, hPrinter,
		  pPrinterName,
		  pi2->pDevMode, pi2->pDevMode,
		  DM_IN_BUFFER | DM_OUT_BUFFER);

	if (lFlag != IDOK)
	{
		GlobalFree(pi2);
		ClosePrinter(hPrinter);
		if (pDevMode)
			GlobalFree(pDevMode);
		return FALSE;
	}

	// Update printer information...
	bFlag = SetPrinter(hPrinter, 2, (LPBYTE)pi2, 0);

    // Note that this is triggered on network printers, although the values
    // get set, so we will ignore this for now
/*	if (!bFlag)
	// The driver doesn't support, or it is unable to make the change...
	{
		GlobalFree(pi2);
		ClosePrinter(hPrinter);
		if (pDevMode)
			GlobalFree(pDevMode);
		return FALSE;
	}
*/
	// Tell other apps that there was a change...
	SendMessageTimeout(HWND_BROADCAST, WM_DEVMODECHANGE, 0L,
			  (LPARAM)(LPCSTR)pPrinterName,
			  SMTO_NORMAL, 1000, NULL);

	// Clean up...
	if (pi2)
		GlobalFree(pi2);
	if (hPrinter)
		ClosePrinter(hPrinter);
	if (pDevMode)
		GlobalFree(pDevMode);

	return TRUE;
}

DWORD GetDefaultPrinterName(LPTSTR PrinterName, DWORD BufferSize)
{
  OSVERSIONINFO osvi;
  DWORD Needed, Returned, rc, BuffSize;
  LPPRINTER_INFO_5 PrinterInfo;
  char* p;

  osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  GetVersionEx(&osvi);

  if (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
  {
    if (!EnumPrinters(PRINTER_ENUM_DEFAULT,NULL,5,NULL,0,&Needed,&Returned))
    {
      if ((rc = GetLastError()) != ERROR_INSUFFICIENT_BUFFER)
        return rc;
    }

    if ((PrinterInfo = LocalAlloc(LPTR,Needed)) == NULL)
      rc = GetLastError();
    else
    {
      if (!EnumPrinters(PRINTER_ENUM_DEFAULT,NULL,5,(LPBYTE) PrinterInfo,Needed,&Needed,&Returned))
        rc = GetLastError();
      else
      {
        if (Returned > 0)
        {
          if ((DWORD) lstrlen(PrinterInfo->pPrinterName) > BufferSize-1)
            rc = ERROR_INSUFFICIENT_BUFFER;
          else
          {
            lstrcpy(PrinterName,PrinterInfo->pPrinterName);
            rc = ERROR_SUCCESS;
          }
        }
        else
        {
          *PrinterName = '0';
          rc = ERROR_SUCCESS;
        }
      }

      LocalFree(PrinterInfo);
    }
  }
  else if (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT)
  {
    if (osvi.dwMajorVersion >= 5) /* Windows 2000 or later */
    {
      BuffSize = BufferSize;
      if (!GetDefaultPrinter(PrinterName,&BuffSize))
        rc = GetLastError();
      else
        rc = ERROR_SUCCESS;
    }
    else /* Windows NT 4.0 or earlier */
    {
      if (GetProfileString("windows","device","",PrinterName,BufferSize) == BufferSize-1)
        return ERROR_INSUFFICIENT_BUFFER;

      p = PrinterName;
      while (*p != '0' && *p != ',')
        ++p;
      *p = '0';

      rc = ERROR_SUCCESS;
    }
  }

  return rc;
}		

LPDEVMODE GetPrinterDevMode(char *pDevice)
{

   HANDLE      hPrinter;
   LPDEVMODE   pDevMode;
   DWORD       dwNeeded, dwRet;

   /* Start by opening the printer */
   if (!OpenPrinter(pDevice, &hPrinter, NULL))
       return NULL;

   /*
    * Step 1:
    * Allocate a buffer of the correct size.
    */
   dwNeeded = DocumentProperties(0,
       hPrinter,       /* Handle to our printer. */
       pDevice,        /* Name of the printer. */
       NULL,           /* Asking for size, so */
       NULL,           /* these are not used. */
       0);             /* Zero returns buffer size. */
   pDevMode = (LPDEVMODE)malloc(dwNeeded);

   /*
    * Step 2:
    * Get the default DevMode for the printer
    */
   dwRet = DocumentProperties(0,
       hPrinter,
       pDevice,
       pDevMode,       /* The address of the buffer to fill. */
       NULL,           /* Not using the input buffer. */
       DM_OUT_BUFFER); /* Have the output buffer filled. */
   if (dwRet != IDOK)
   {
       /* If failure, cleanup and return failure. */
       free(pDevMode);
       ClosePrinter(hPrinter);
       return NULL;
   }

   /* Finished with the printer */
   ClosePrinter(hPrinter);

   /* Return the DevMode structure. */
   return pDevMode;
}
				

